// 12th Of August 2008
/*
 * Permission is given to modify this code for the purpose
 * of enhancing Richard Rognlie's PBMSERV, either by adding new
 * functionality to the Oshi module, or as a base for a new game.  If you
 * wish to use the code for any other purposes, ASK!  
*/

/* 
   Author: Martin Moller Skarbiniks Pedersen
   Userid: chaz 
   Email:  traxplayer@gmail.com
*/
 
#include <sys/types.h>
#include <cctype>
#include <ctime>
#include <cstdlib>
#include <cstring>
#include <cassert>
#include "oshi.h"
#include "sendmail.h"
#include "strutl.h"


unsigned int Oshi::WhiteTiersCaptured() 
{
  unsigned int result=14;
  int i,j;

  for (i=MinRow(); i<=MaxRow()/3; i++) {
    for (j=MinCol(); j<=MaxCol(); j++) {
      switch (ToTier(GetAt(i,j))) {
      case 0: 
	break;
      case 1:
	if (Colour(i,j)=='W') result--;
	break;
      case 2:
	if (Colour(i,j)=='W') result-=2;
	break;
      case 3:
	if (Colour(i,j)=='W') result-=3;
	break;
      default:
	assert(0); // This should never happen
      }
    }
  }
  return result;
}


const bool Oshi::InsideBoard(const int &row, const int &col) const 
{
  if (row<0) return false;
  if (row>8) return false;
  if (col<0) return false;
  if (col>8) return false;
  return true;
}


char * Oshi::Push(const int &row,const int &col,const int &dir_row,const int &dir_col)
{
  // returns pieces which are pushed and must be checked for their position
  int end_row=row;
  int end_col=col;
  char *result;
  unsigned int result_count=0;
  char save1,save2;

  result=(char *)malloc(5);
  save2=Blank();
  while ((InsideBoard(end_row,end_col)) && GetAt(end_row,end_col)!=Blank()) {
    result[result_count]=GetAt(end_row,end_col);
    result_count++;
    save1=GetAt(end_row,end_col);
    PutAt(end_row,end_col,save2);
    save2=save1;
    end_row+=dir_row;
    end_col+=dir_col;
  }
  if (InsideBoard(end_row,end_col)) PutAt(end_row,end_col,save2);
  result[result_count]='\0';

  return result;
}
    
  
unsigned int Oshi::HowManyPushed(int row, int col,int dir_row,int dir_col,const int &spaces) 
{
  /*
    returns how many pieces affected if (row,col)
    is pushed in a direction
  */
  //fprintf(stderr,"Oshi::HowManyPushed(int %d,int %d,int %d,int %d,int %d) called.\n",row,col,dir_row,dir_col,spaces);
  
  int direction_row=0, direction_col=0;
  unsigned int i,result=0;

  assert(spaces>0);
  assert(spaces<=3);
  row+=dir_row;
  col+=dir_col;
  while (InsideBoard(row,col) && (GetAt(row,col)!=Blank())) {
    result++;
    row+=dir_row;
    col+=dir_col;
  } 
  //fprintf(stderr,"result: %d\n",result);
  return result;
}


unsigned int Oshi::RedTiersCaptured() 
{
  unsigned int result=14;
  int i,j;

  for (i=MinRow(); i<=MaxRow()/3; i++) {
    for (j=MinCol(); j<=MaxCol(); j++) {
      switch (ToTier(GetAt(i,j))) {
      case 0: 
	break;
      case 1:
	if (Colour(i,j)=='R') result--;
	break;
      case 2:
	if (Colour(i,j)=='R') result-=2;
	break;
      case 3:
	if (Colour(i,j)=='R') result-=3;
	break;
      default:
	assert(0); // This should never happen
      }
    }
  }
  return result;
}


void Oshi::PrintBoard(FILE *fp)
{
  unsigned int red_tiers=0;
  unsigned int white_tiers=0;
  unsigned int r_captured,w_captured;
  int i,j;
  int num[4];
  
  w_captured=WhiteTiersCaptured();
  r_captured=RedTiersCaptured();

  fprintf(fp,"      ");
  for (j=MinCol(); j<=MaxCol(); j++)
    fprintf(fp,"%c   ",j+'A');
  fprintf(fp,"\n");
  for (i=MaxRow()/3 ; i>=MinRow() ; i--) {
    fprintf(fp,"    +---+---+---+---+---+---+---+---+---+\n");
    fprintf(fp,"  %d |",i+1);
    for (j=MinCol(); j<=MaxCol(); j++) {
      switch (ToTier(GetAt(i,j))) {
      case 0:
	fprintf(fp,"   |");
	break;
      case 1:
	if (Colour(i,j)=='W')
	  fprintf(fp,"*1*|");
	else
	  fprintf(fp,":1:|");
	break;
      case 2:
	if (Colour(i,j)=='W')
	  fprintf(fp,"*2*|");
	else
	  fprintf(fp,":2:|");
	break;
      case 3:
	if (Colour(i,j)=='W')
	  fprintf(fp,"*3*|");
	else
	  fprintf(fp,":3:|");
	break;
      default:
	assert(0); // This should never happen
	break;
      }
    }
    fprintf(fp,"  %d",i+1);
    if (i==4) {
      fprintf(fp,"    Red tiers captured: %d",r_captured);
    }
    if (i==5) {
      fprintf(fp,"  White tiers captured: %d",w_captured);
    }
    fprintf(fp,"\n");
  }
  fprintf(fp,"    +---+---+---+---+---+---+---+---+---+\n");
  fprintf(fp,"      ");
  for (j=MinCol(); j<=MaxCol(); j++)
    fprintf(fp,"%c   ",j+'A');
  fprintf(fp,"\n");
}


int Oshi::IsGameOver(const char *&winner) 
{
  if ((WhiteTiersCaptured()>=7) && (RedTiersCaptured()>=7)) {
    // DRAW
    winner=NULL;
    return 1;
  }
  if ((WhiteTiersCaptured()>=7) && (RedTiersCaptured()<7)) {
    // Red wins
    winner=players[0]; 
    return 1;
  }
  if ((WhiteTiersCaptured()<7) && (RedTiersCaptured()>=7)) {
    // White wins
    winner=players[1];
    return 1;
  }
  // No winner yet. Game continue
  return 0; 
}


char Oshi::Colour(const char &c) 
{
  // Returns 'R'ed, 'W'hite or 'N'one
  if (Blank()==c) return 'N';
  if ((c>='A') && (c<='H')) return 'R';
  return 'W';
}

char Oshi::Colour(const int &x,const int &y) 
{
  // Returns 'R'ed, 'W'hite or 'N'one
  return Colour(GetAt(x,y));
}


int Oshi::ToTier(const int &row,const int &col) 
{
  // Returns the tier at (row,col) or 0 if no piece on (row,col)
  return ToTier(GetAt(row,col));
}


int Oshi::ToTier(const char &c) 
{
  // Returns the tier or 0 if no piece */
  if (c==Blank()) return 0; // Empty
  switch (c) {
  case 'A':
  case 'B':
  case 'C':
  case 'D':
  case 'I':
  case 'J':
  case 'K':
  case 'L':
    return 1;
  case 'E':
  case 'F':
  case 'M':
  case 'N':
    return 2;
  case 'G':
  case 'H':
  case 'Q':
  case 'P':
    return 3;
  default:
    assert(0); // This should never happen
  }
  assert(0); // This should never happen
  return 0; // To make the compiler happy
}

int Oshi::Init(void) {
/* The start position
 *3* is a White (player2) 3-tier
 :3: is a Red   (player1) 3-tier

      A   B   C   D   E   F   G   H   I
    +---+---+---+---+---+---+---+---+---+
  9 |*3*|   |   |   |   |   |   |   |*3*|  9 
    +---+---+---+---+---+---+---+---+---+ 
  8 |   |   |*2*|*1*|*1*|*1*|*2*|   |   |  8 
    +---+---+---+---+---+---+---+---+---+
  7 |   |   |   |   |*1*|   |   |   |   |  7 
    +---+---+---+---+---+---+---+---+---+
  6 |   |   |   |   |   |   |   |   |   |  6 
    +---+---+---+---+---+---+---+---+---+
  5 |   |   |   |   |   |   |   |   |   |  5 
    +---+---+---+---+---+---+---+---+---+
  4 |   |   |   |   |   |   |   |   |   |  4 
    +---+---+---+---+---+---+---+---+---+
  3 |   |   |   |   |:1:|   |   |   |   |  3 
    +---+---+---+---+---+---+---+---+---+
  2 |   |   |:2:|:1:|:1:|:1:|:2:|   |   |  2 
    +---+---+---+---+---+---+---+---+---+  
  1 |:3:|   |   |   |   |   |   |   |:3:|  1 
    +---+---+---+---+---+---+---+---+---+
      A   B   C   D   E   F   G   H   I
*/

  int i,j;

  Board2D::Init(27, 9);

  for (i=MinRow(); i<=MaxRow()/3; i++)
    for (j=MinCol(); j<=MaxCol(); j++)
      PutAt(i,j,Blank());
  
  /* 
     Red   1-tier A,B,C and D
     Red   2-tier E and F
     Red   3-tier G and H
     White 1-tier I,J,K and L
     White 2-tier M and N
     White 3-tier Q and P
  */
  
  // First the Red piece
  PutAt(0,0,'G');
  PutAt(0,8,'H');
  PutAt(1,2,'E');
  PutAt(1,3,'A');
  PutAt(1,4,'B');
  PutAt(1,5,'C');
  PutAt(1,6,'F');
  PutAt(2,4,'D');
  
  // Then the White piece  
  PutAt(8,0,'Q');
  PutAt(8,8,'P');
  PutAt(7,2,'M');
  PutAt(7,3,'I');
  PutAt(7,4,'J');
  PutAt(7,5,'K');
  PutAt(7,6,'N');
  PutAt(6,4,'L');
  
  return 1;
}


const bool Oshi::CheckPiece(const char &c) {
  unsigned int i,j;
  unsigned int row=-1;
  unsigned int col=-1;

  fprintf(stderr,"CheckPiece(%c) called.\n",c);
  for (i=0; i<=MaxRow()/3; i++) {
    for (j=0; j<=MaxCol(); j++) {
      if (GetAt(i,j)==c) {
	row=i;
	col=j;
      }
    }
  }
  if (row>=0) {
    fprintf(stderr,"row=%d, GetAt(%d,%d)=%c\n",row,1+row+MaxRow()/3,col,GetAt(1+row+MaxRow()/3,col));
  }
  if ((row>=0) && (GetAt(1+row+MaxRow()/3,col)==c)) return false;
  return true;
}

    

void Oshi::CopyBoardToStorage1()
{
  int i,j;
  for (i=0; i<=MaxRow()/3; i++) {
    for (j=0; j<=MaxCol(); j++) {
      PutAt(1+i+MaxRow()/3,j,GetAt(i,j));
    }
  }
}


void Oshi::CopyBoardToStorage2()
{
  int i,j;
  for (i=0; i<=MaxRow()/3; i++) {
    for (j=0; j<=MaxCol(); j++) {
      PutAt(1+i+2*MaxRow()/3,j,GetAt(i,j));
    }
  }
}


void Oshi::CopyStorage1ToBoard()
{
  int i,j;
  for (i=0; i<=MaxRow()/3; i++) {
    for (j=0; j<=MaxCol(); j++) {
      PutAt(i,j,GetAt(1+i+MaxRow()/3,j));
    }
  }
}


void Oshi::CopyStorage2ToBoard()
{
  int i,j;
  for (i=0; i<=MaxRow()/3; i++) {
    for (j=0; j<=MaxCol(); j++) {
      PutAt(i,j,GetAt(1+i+2*MaxRow()/3,j));
    }
  }
}


void Oshi::CopyStorage2ToStorage1()
{
  int i,j;
  for (i=0; i<=MaxRow()/3; i++) {
    for (j=0; j<=MaxCol(); j++) {
      PutAt(1+i+MaxRow()/3,j,GetAt(1+i+2*MaxRow()/3,j));
    }
  }
}
    

const char *Oshi::MakeMove(const char *move)
{
  int fromrow, fromcol, torow, tocol;
  int direction_row=0,direction_col=0;
  unsigned int r,c,i;
  int WhoToMove=CurrentPlayer()+1; /* wtm: 1=Player 1, 2=Player 2 */ 
  static char newmove[5];
  unsigned int move_length;
  char *pushed_pieces=NULL;

  /* first decode the move and check for simple errors */
  
  pushed_pieces=(char *)malloc(20); // 10 should be enough
  pushed_pieces[0]='\0';
  if (strlen(move)<4 || strlen(move)>5) {
    return (char*)Error("Please enter move in the format A1-A4");
  }
  newmove[0]=toupper(move[0]);
  newmove[1]=toupper(move[1]);
  if (strlen(move)==5) {
    newmove[2]=toupper(move[3]);
    newmove[3]=toupper(move[4]);
  }
  else {
    newmove[2]=toupper(move[2]);
    newmove[3]=toupper(move[3]);
  }
  newmove[4]='\0';
  fromcol=newmove[0]-'A';
  fromrow=newmove[1]-'1';
  tocol  =newmove[2]-'A';
  torow  =newmove[3]-'1';
  //fprintf(stderr,"Colour(%d,%d)=%c\n",fromrow,fromcol,Colour(fromrow,fromcol));
  if ((fromrow<0) || (fromrow>8)) {
    return (char*)Error("Illegal row: %c",move[0]);
  }
  if ((fromcol<0) || (fromcol>8)) {
    return (char*)Error("Illegal column: %c",move[1]);
  }
  if ((torow<0) || (torow>8)) {
    return (char*)Error("Illegal row: %c",move[2]);
  }
  if ((tocol<0) || (tocol>8)) {
    return (char*)Error("Illegal column: %c",move[3]);
  }
  if ((WhoToMove==1) && (Colour(fromrow,fromcol)!='R')) {
    return (char*)Error("Not your piece");
  }
  if ((WhoToMove==2) && (Colour(fromrow,fromcol)!='W')) {
    return (char*)Error("Not your piece");
  }
  if (fromrow<torow) direction_row= 1;
  if (fromrow>torow) direction_row=-1;
  if (fromcol<tocol) direction_col= 1;
  if (fromcol>tocol) direction_col=-1;
  if (direction_row*direction_col!=0) {
    return (char*)Error("Illegal move. You can't move diagonal");
  }
  if (direction_row+direction_col==0) {
    return (char*)Error("Illegal move. You must move a piece");
  }
  move_length=abs(fromrow-torow)+abs(fromcol-tocol);
  if (move_length>ToTier(fromrow,fromcol)) {
    return (char*)Error("Illegal move. The piece on %c%c can maximum move %d spaces.",newmove[0],newmove[1],ToTier(fromrow,fromcol));
  }
  if (HowManyPushed(fromrow,fromcol,direction_row,direction_col,move_length)>ToTier(fromrow,fromcol)) {
    return (char*)Error("Illegal move. The piece on %c%c isn't strong enough to push %d pieces",newmove[0],newmove[1],HowManyPushed(fromrow,fromcol,direction_row,direction_col,move_length));
  }
  // save a copy if the move is illegal so the original board can be restored
  CopyBoardToStorage2(); 
  r=fromrow;
  c=fromcol;
  for (i=0; i<move_length; i++) {
    sprintf(pushed_pieces,"%s%s",pushed_pieces,Push(r,c,direction_row,direction_col));
    r+=direction_row;
    c+=direction_col;
  }
  for (i=0; i<strlen(pushed_pieces); i++) {
    if (!CheckPiece(pushed_pieces[i])) {
      CopyStorage2ToBoard();
      return (char*)Error("Illegal move. You can't push a piece back to it previous position.");
    }
  }
  /* move is ok. */
  CopyStorage2ToStorage1();

  return newmove;

}


const char *Oshi::PlayerNames(int n) { return "Reds (:)\0Whites (*)"+n*9; }
  
  
